import os
import struct
import marshal
import zlib
import sys
import shutil
# import imp
import types
from uuid import uuid4 as uniquename

cwd = os.getcwd()

class CTOCEntry:
    def __init__(self, position, entrySize, entryPos, cmprsdDataSize, uncmprsdDataSize, cmprsFlag, typeCmprsData, name, titlebytes):
        self.position = position
        self.entrySize = entrySize
        self.entryPos = entryPos
        self.cmprsdDataSize = cmprsdDataSize
        self.uncmprsdDataSize = uncmprsdDataSize
        self.cmprsFlag = cmprsFlag
        self.typeCmprsData = typeCmprsData
        self.name = name
        self.titlebytes = titlebytes
    def __repr__(self):
        d = {}
        d['entrySize'] = self.entrySize
        d['entryPos'] = self.entryPos
        d['position'] = self.position
        d['cmprsdDataSize'] = self.cmprsdDataSize
        d['uncmprsdDataSize'] = self.uncmprsdDataSize
        d['cmprsFlag'] = self.cmprsFlag
        d['typeCmprsData'] = self.typeCmprsData
        d['name'] = self.name
        d['titlebytes'] = self.titlebytes
        return str(d)
    def to_bytes(self):
        r = struct.pack('!iiiiBc', 
            self.entrySize,
            self.entryPos,
            self.cmprsdDataSize,
            self.uncmprsdDataSize,
            self.cmprsFlag,
            self.typeCmprsData,
        ) + self.name.encode()
        return r + b'\x00' * (self.entrySize - len(r))

class PyTitle():
    def __init__(self, x):
        self.pyinstVer = x.pyinstVer
        if x.pyinstVer == 20:
            self.bytes = x.fPtr[x.py_start_20:x.py_start_20+x.PYINST20_COOKIE_SIZE]
            (magic, lengthofPackage, toc, tocLen, pyver) = struct.unpack('!8siiii', self.bytes)
        elif x.pyinstVer == 21:
            self.bytes = x.fPtr[x.py_start_21:x.py_start_21+x.PYINST21_COOKIE_SIZE]
            (magic, lengthofPackage, toc, tocLen, pyver, pylibname) = struct.unpack('!8siiii64s', self.bytes)
        else:
            raise Exception('not good pyinstver')
        self.magic = magic
        self.lengthofPackage = lengthofPackage
        self.toc = toc
        self.tocLen = tocLen
        self.pyver = pyver
        self.pylibname = pylibname
    def to_bytes(self):
        if self.pyinstVer == 20:
            r = struct.pack('!8siiii', self.magic, self.lengthofPackage, self.toc, self.tocLen, self.pyver)
        elif self.pyinstVer == 21:
            r = struct.pack('!8siiii64s', self.magic, self.lengthofPackage, self.toc, self.tocLen, self.pyver, self.pylibname)
        return r
class PyInstArchive:
    PYINST20_COOKIE_SIZE = 24           # For pyinstaller 2.0
    PYINST21_COOKIE_SIZE = 24 + 64      # For pyinstaller 2.1+
    MAGIC = b'MEI\014\013\012\013\016'  # Magic number which identifies pyinstaller
    def __init__(self, path):
        self.filePath = path
    def open(self):
        self.fPtr = open(self.filePath, 'rb').read()
        self.fileSize = len(self.fPtr)
        self.py_start_20 = self.fileSize - self.PYINST20_COOKIE_SIZE
        self.py_start_21 = self.fileSize - self.PYINST21_COOKIE_SIZE
        return True
    def checkFile(self):
        print('[*] Processing {0}'.format(self.filePath))
        magicFromFile = self.fPtr[self.py_start_20:self.py_start_20+len(self.MAGIC)]
        if magicFromFile == self.MAGIC:
            self.pyinstVer = 20
            print('[*] Pyinstaller version: 2.0')
            return True
        magicFromFile = self.fPtr[self.py_start_21:self.py_start_21+len(self.MAGIC)]
        if magicFromFile == self.MAGIC:
            print('[*] Pyinstaller version: 2.1+')
            self.pyinstVer = 21
            return True
        print('[*] Error : Unsupported pyinstaller version or not a pyinstaller archive')
        return False
    def getCArchiveInfo(self):
        titler = PyTitle(self)
        magic = titler.magic
        lengthofPackage = titler.lengthofPackage
        toc = titler.toc
        tocLen = titler.tocLen
        self.titler = titler
        self.pyver = titler.pyver
        pylibname = titler.pylibname
        self.overlaySize = lengthofPackage
        self.overlayPos = self.fileSize - self.overlaySize
        self.tableOfContentsPos = self.overlayPos + toc
        self.tableOfContentsSize = tocLen
        return True
    def parseTOC(self):
        # Go to the table of contents
        start_table = self.tableOfContentsPos
        self.tocList = []
        parsedLen = 0
        def read_n(n):
            nonlocal start_table
            r = self.fPtr[start_table:start_table+n]
            start_table += n
            return r
        # Parse table of contents
        while parsedLen < self.tableOfContentsSize:
            bys = read_n(4)
            (entrySize, ) = struct.unpack('!i', bys)
            nameLen = struct.calcsize('!iiiiBc')
            bys2 = read_n(entrySize - 4)
            (entryPos, cmprsdDataSize, uncmprsdDataSize, cmprsFlag, typeCmprsData, name) = \
            struct.unpack('!iiiBc{0}s'.format(entrySize - nameLen), bys2)
            titlebytes = bys + bys2
            # print(entryPos, entrySize, titlebytes, cmprsdDataSize)
            name = name.decode('utf-8').rstrip('\0')
            if len(name) == 0:
                name = str(uniquename())
                print('[!] Warning: Found an unamed file in CArchive. Using random name {0}'.format(name))
            self.tocList.append(
                CTOCEntry(
                    self.overlayPos + entryPos,
                    entrySize,
                    entryPos,
                    cmprsdDataSize,
                    uncmprsdDataSize,
                    cmprsFlag,
                    typeCmprsData,
                    name,
                    titlebytes
                ))
            parsedLen += entrySize
        print('[*] Found {0} files in CArchive'.format(len(self.tocList)))
    def extractFiles(self, exefile=None):
        print('[*] Beginning extraction...please standby')
        extractionDir = os.path.join(cwd, os.path.basename(self.filePath) + '_extracted')
        if not os.path.exists(extractionDir):
            os.mkdir(extractionDir)
        os.chdir(extractionDir)

        diff = 0
        if exefile:
            fPtrcp = self.fPtr[:]
        for entry in self.tocList:
            basePath = os.path.dirname(entry.name)
            if basePath != '':
                if not os.path.exists(basePath):
                    os.makedirs(basePath)
            data = obytes = self.fPtr[entry.position:entry.position+entry.cmprsdDataSize]
            if entry.cmprsFlag == 1:
                data = zlib.decompress(data)
                assert len(data) == entry.uncmprsdDataSize # Sanity Check
            if exefile:
                if diff:
                    t1 = entry.to_bytes()
                    entry.entryPos += diff
                    t2 = entry.to_bytes()
                    assert len(fPtrcp.split(t1)) == 2
                    fPtrcp = fPtrcp.replace(t1, t2, 1)
                if os.path.isfile(entry.name):
                    with open(entry.name, 'rb') as f:
                        cbytes = f.read()
                    if cbytes != data:
                        print('[*] change file:', entry.name)
                        newolen = len(cbytes)
                        cbytes = zlib.compress(cbytes)
                        newclen = len(cbytes)
                        fPtrcp = fPtrcp.replace(obytes, cbytes, 1)
                        t1 = entry.to_bytes()
                        entry.cmprsdDataSize = newclen
                        entry.uncmprsdDataSize = newolen
                        t2 = entry.to_bytes()
                        fPtrcp = fPtrcp.replace(t1, t2)
                        diff += len(fPtrcp) - len(self.fPtr)
                else:
                    with open(entry.name, 'wb') as f:
                        f.write(data)
            else:
                with open(entry.name, 'wb') as f:
                    f.write(data)
                if entry.typeCmprsData == b'z':
                    self._extractPyz(entry.name)
        if exefile:
            t1 = self.titler.to_bytes()
            self.titler.lengthofPackage += diff
            self.titler.toc += diff
            t2 = self.titler.to_bytes()
            fPtrcp = fPtrcp.replace(t1, t2)
            with open(os.path.join(cwd, exefile), 'wb') as f:
                f.write(fPtrcp)
        os.chdir(cwd)
    def _extractPyz(self, name):
        dirName =  name + '_extracted'
        # Create a directory for the contents of the pyz
        if not os.path.exists(dirName):
            os.mkdir(dirName)
        with open(name, 'rb') as f:
            pyzMagic = f.read(4)
            assert pyzMagic == b'PYZ\0' # Sanity Check
            pycHeader = f.read(4) # Python magic value
            # if imp.get_magic() != pycHeader:
            #     print('[!] Warning: The script is running in a different python version than the one used to build the executable')
            #     print('    Run this script in Python{0} to prevent extraction errors(if any) during unmarshalling'.format(self.pyver))
            (tocPosition, ) = struct.unpack('!i', f.read(4))
            f.seek(tocPosition, os.SEEK_SET)
            try:
                toc = marshal.load(f)
            except:
                print('[!] Unmarshalling FAILED. Cannot extract {0}. Extracting remaining files.'.format(name))
                return
            print('[*] Found {0} files in PYZ archive'.format(len(toc)))
            if type(toc) == list:
                toc = dict(toc)
            for key in toc.keys():
                (ispkg, pos, length) = toc[key]
                f.seek(pos, os.SEEK_SET)
                fileName = key
                try:
                    fileName = key.decode('utf-8')
                except:
                    pass
                destName = os.path.join(dirName, fileName.replace("..", "__"))
                destDirName = os.path.dirname(destName)
                if not os.path.exists(destDirName):
                    os.makedirs(destDirName)
                try:
                    data = f.read(length)
                    data = zlib.decompress(data)
                except:
                    print('[!] Error: Failed to decompress {0}, probably encrypted. Extracting as is.'.format(fileName))
                    open(destName + '.pyc.encrypted', 'wb').write(data)
                    continue
                with open(destName + '.pyc', 'wb') as pycFile:
                    pycFile.write(pycHeader)      # Write pyc magic
                    pycFile.write(b'\0' * 4)      # Write timestamp
                    if self.pyver >= 33:
                        pycFile.write(b'\0' * 4)  # Size parameter added in Python 3.3
                    pycFile.write(data)
def extract_all(exe_path, exefile=None):
    arch = PyInstArchive(exe_path)
    if arch.open():
        if arch.checkFile():
            if arch.getCArchiveInfo():
                arch.parseTOC()
                arch.extractFiles(exefile=exefile)
                print('[*] Successfully extracted pyinstaller archive: {0}'.format(exe_path))
                print('')
                print('You can now use a python decompiler on the pyc files within the extracted directory')
                return
        arch.close()
def main():
    if len(sys.argv) < 2:
        print('[*] Usage: pyinstxtractor.py <filename>')
    else:
        extract_all(sys.argv[1])
if __name__ == '__main__':
    extract_all('./labelImg.exe')
    with open('./labelImg', encoding='utf-8') as f:
        with open('./labelImg.exe_extracted/labelImg', 'w', encoding='utf-8') as f2:
            f2.write(f.read())
    extract_all('./labelImg.exe', '../newLabelImg.exe')
    if not os.path.isfile('../labelimg.cfg'):
        with open('../labelimg.cfg', 'w', encoding='utf-8') as f:
            f.write(('''
defaultFilename=data
defaultSaveDir=data
            ''').strip())
    shutil.rmtree('./labelImg.exe_extracted')

    # extract_all('./labelImg.exe', False)
    # extract_all('./new.exe', False)
